import * as ast from "@babel/types";
import { UnsupportedBase } from "./unsupported";
export declare class FreeVariables extends UnsupportedBase {
    private scope;
    private defineMode;
    /**
     * A short hand of `visit()` and `variables()`.
     * @param root the node to analyze
     * @returns a list of free variable names
     */
    analyze(root: ast.Node): string[];
    /**
     * Get all free variables as a copy.
     * @returns all free variables
     */
    getResult(): Set<string>;
    private enter;
    private leave;
    /**
     * When in pattern matching nodes, we set `isDefineMode` to true.
     * So that `identifier` will define new symbols rather than resolve them.
     */
    private get isDefineMode();
    private setDefineMode;
    private unsetDefineMode;
    /**
     * Defer a visit to the end of the scope.
     * Usually used for function scopes and class scopes.
     * For example, the variable `count` is not free in following code.
     *
     * ```javascript
     * function counter() {
     *   function increment() {
     *     count += 1;
     *   }
     *   let count = 0;
     *   return increment;
     * }
     * ```
     * @param visitor the visitor function
     */
    private deferVisit;
    /**
     * Visit a syntax tree node.
     * @param node the node to visit
     * @returns nothing
     */
    visit(node: ast.Node): void;
    private visitAll;
    visitNullable(node: ast.Node | null | undefined): void;
    /**
     * Used for `XStatement` and `ArrowFunctionExpression`.
     * Because calling `this.blockStatement` will create a scope.
     * If the statement is a `BlockStatement`, it will not create a scope.
     * @param body `body` of `Statement` or `ArrowFunctionExpression`.
     */
    private body;
    private file;
    private program;
    private blockStatement;
    private breakStatement;
    private continueStatement;
    private debuggerStatement;
    private doWhileStatement;
    private emptyStatement;
    private expressionStatement;
    private forInStatement;
    private forStatement;
    private functionDeclaration;
    private ifStatement;
    private labeledStatement;
    private returnStatement;
    private switchStatement;
    private throwStatement;
    private tryStatement;
    private variableDeclaration;
    private whileStatement;
    private withStatement;
    private classDeclaration;
    private exportAllDeclaration;
    private exportDefaultDeclaration;
    private exportNamedDeclaration;
    private forOfStatement;
    private importDeclaration;
    private arrayExpression;
    private assignmentExpression;
    private binaryExpression;
    private callExpression;
    private conditionalExpression;
    private functionExpression;
    private identifier;
    private stringLiteral;
    private numericLiteral;
    private nullLiteral;
    private booleanLiteral;
    private regExpLiteral;
    private logicalExpression;
    private memberExpression;
    private newExpression;
    private objectExpression;
    private sequenceExpression;
    private parenthesizedExpression;
    private thisExpression;
    private unaryExpression;
    private updateExpression;
    arrowFunctionExpression(arrowFunctionExpression: ast.ArrowFunctionExpression): void;
    classExpression(classExpression: ast.ClassExpression): void;
    metaProperty(metaProperty: ast.MetaProperty): void;
    private super;
    taggedTemplateExpression(taggedTemplateExpression: ast.TaggedTemplateExpression): void;
    templateLiteral(templateLiteral: ast.TemplateLiteral): void;
    templateElement(templateElement: ast.TemplateElement): void;
    yieldExpression(yieldExpression: ast.YieldExpression): void;
    awaitExpression(awaitExpression: ast.AwaitExpression): void;
    import(_import: ast.Import): void;
    bigIntLiteral(_bigIntLiteral: ast.BigIntLiteral): void;
    optionalMemberExpression(optionalMemberExpression: ast.OptionalMemberExpression): void;
    optionalCallExpression(optionalCallExpression: ast.OptionalCallExpression): void;
    jsxIdentifier(_jsxIdentifier: ast.JSXIdentifier): void;
    jsxMemberExpression(jsxMemberExpression: ast.JSXMemberExpression): void;
    jsxElement(jsxElement: ast.JSXElement): void;
    jsxSpreadChild(jsxSpreadChild: ast.JSXSpreadChild): void;
    jsxFragment(jsxFragment: ast.JSXFragment): void;
    jsxOpeningElement(jsxOpeningElement: ast.JSXOpeningElement): void;
    jsxOpeningFragment(_jsxOpeningFragment: ast.JSXOpeningFragment): void;
    jsxClosingElement(_jsxClosingElement: ast.JSXClosingElement): void;
    jsxClosingFragment(_jsxClosingFragment: ast.JSXClosingFragment): void;
    jsxAttribute(jsxAttribute: ast.JSXAttribute): void;
    jsxSpreadAttribute(jsxSpreadAttribute: ast.JSXSpreadAttribute): void;
    private jsxExpressionContainer;
    private jsxText;
    private jsxEmptyExpression;
    private bindExpression;
    private pipelinePrimaryTopicReference;
    private doExpression;
    private recordExpression;
    private tupleExpression;
    private decimalLiteral;
    private moduleExpression;
    private restElement;
    private assignmentPattern;
    private arrayPattern;
    private objectPattern;
    private objectMethod;
    private objectProperty;
    private catchClause;
    /**
     * Let’s discuss the type of `declarator.id` is `LVal`.
     *
     * - Identifier: e.g. `y` in `let x = y`.
     * - MemberExpression: impossible.
     * - RestElement: impossible.
     * - AssignmentPattern: impossible.
     * - ArrayPattern: e.g. let [x, ..xs] = a;
     * - ObjectPattern: e.g. let { x, y = z, y: { t } } = a;
     * - TSParameterProperty: unsupported.
     *
     * @param declarator the `VariableDeclarator` to visit
     */
    private variableDeclarator;
    /**
     * E.g. `call(?, test)`.
     * @param _argumentPlaceholder the placeholder operator `?`
     */
    private argumentPlaceholder;
    private jsxNamespacedName;
    private spreadElement;
    private v8IntrinsicIdentifier;
    private privateName;
    private noop;
    private classMethod;
    private classBody;
    private classPrivateMethod;
    /**
     * Stage 3: https://github.com/tc39/proposal-class-static-block
     * @param staticBlock the `StaticBlock` to visit.
     */
    private staticBlock;
    private classPrivateProperty;
    private classProperty;
    private decorator;
    private directive;
    private directiveLiteral;
    protected exportSpecifier(_exportSpecifier: ast.ExportSpecifier): void;
    protected importDefaultSpecifier(_importDefaultSpecifier: ast.ImportDefaultSpecifier): void;
    protected importNamespaceSpecifier(_importNamespaceSpecifier: ast.ImportNamespaceSpecifier): void;
    protected importSpecifier(_importSpecifier: ast.ImportSpecifier): void;
    protected exportNamespaceSpecifier(_exportNamespaceSpecifier: ast.ExportNamespaceSpecifier): void;
    protected exportDefaultSpecifier(_exportDefaultSpecifier: ast.ExportDefaultSpecifier): void;
    protected importAttribute(_importAttribute: ast.ImportAttribute): void;
    /**
     * See https://github.com/tc39/proposal-hashbang
     * @param _interpreterDirective the node
     */
    protected interpreterDirective(_interpreterDirective: ast.InterpreterDirective): void;
    private pipelineBareFunction;
    private pipelineTopicExpression;
    private placeholder;
    private switchCase;
    /** @deprecated */
    private spreadProperty;
    /** @deprecated */
    private restProperty;
    /** @deprecated */
    private numberLiteral;
    /** @deprecated */
    private regexLiteral;
}
