"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FreeVariables = void 0;
const error_1 = require("./error");
const helpers_1 = require("./helpers");
const scope_1 = require("./scope");
const unsupported_1 = require("./unsupported");
class FreeVariables extends unsupported_1.UnsupportedBase {
    constructor() {
        super(...arguments);
        this.scope = new scope_1.Scope();
        this.defineMode = [];
        // =========================
        // End: some deprecated node
        // =========================
    }
    /**
     * A short hand of `visit()` and `variables()`.
     * @param root the node to analyze
     * @returns a list of free variable names
     */
    analyze(root) {
        this.visit(root);
        return this.scope.unresolvedNames;
    }
    /**
     * Get all free variables as a copy.
     * @returns all free variables
     */
    getResult() {
        return new Set(this.scope.unresolvedNames);
    }
    // =======================
    // Start: shared utilities
    // =======================
    enter(scopable) {
        this.scope = new scope_1.Scope(scopable, this.scope);
    }
    leave(expectedScopable) {
        const { enclosingScope, scopable } = this.scope;
        if (expectedScopable !== scopable) {
            this.reportError(error_1.Error.ScopeMismatch);
        }
        if (enclosingScope) {
            this.scope.runScheduledTasks();
            this.scope.liftUnresolvedNames();
            this.scope = enclosingScope;
        }
    }
    /**
     * When in pattern matching nodes, we set `isDefineMode` to true.
     * So that `identifier` will define new symbols rather than resolve them.
     */
    get isDefineMode() {
        const { length } = this.defineMode;
        return length === 0 ? false : this.defineMode[length - 1];
    }
    setDefineMode(value) {
        this.defineMode.push(value);
    }
    unsetDefineMode() {
        this.defineMode.pop();
    }
    /**
     * Defer a visit to the end of the scope.
     * Usually used for function scopes and class scopes.
     * For example, the variable `count` is not free in following code.
     *
     * ```javascript
     * function counter() {
     *   function increment() {
     *     count += 1;
     *   }
     *   let count = 0;
     *   return increment;
     * }
     * ```
     * @param visitor the visitor function
     */
    deferVisit(visitor) {
        this.scope.scheduleTask(visitor);
    }
    /**
     * Visit a syntax tree node.
     * @param node the node to visit
     * @returns nothing
     */
    visit(node) {
        this[helpers_1.uncapitalize(node.type)](node);
    }
    visitAll(nodes) {
        return nodes.forEach(this.visit.bind(this));
    }
    visitNullable(node) {
        if (node) {
            this.visit(node);
        }
    }
    /**
     * Used for `XStatement` and `ArrowFunctionExpression`.
     * Because calling `this.blockStatement` will create a scope.
     * If the statement is a `BlockStatement`, it will not create a scope.
     * @param body `body` of `Statement` or `ArrowFunctionExpression`.
     */
    body(body) {
        if (body.type === "BlockStatement") {
            this.blockStatement(body, false);
        }
        else {
            this.visit(body);
        }
    }
    // =====================
    // End: shared utilities
    // =====================
    file(file) {
        return this.program(file.program);
    }
    program(program) {
        return this.visitAll(program.body);
    }
    // ============================================
    // Start: all possible sub-types of `Statement`
    // ============================================
    blockStatement(blockStatement, scopic = true) {
        if (scopic) {
            this.enter(blockStatement);
        }
        this.visitAll(blockStatement.body);
        if (scopic) {
            this.leave(blockStatement);
        }
    }
    breakStatement(_breakStatement) { }
    continueStatement(_continueStatement) { }
    debuggerStatement(_debuggerStatement) { }
    doWhileStatement(doWhileStatement) {
        this.visit(doWhileStatement.test);
        this.body(doWhileStatement.body);
    }
    emptyStatement(_emptyStatement) { }
    expressionStatement(expressionStatement) {
        this.visit(expressionStatement.expression);
    }
    forInStatement(forInStatement) {
        this.enter(forInStatement);
        this.visit(forInStatement.left);
        this.visit(forInStatement.right);
        this.body(forInStatement.body);
        this.leave(forInStatement);
    }
    forStatement(forStatement) {
        this.enter(forStatement);
        this.visitNullable(forStatement.init);
        this.visitNullable(forStatement.test);
        this.visitNullable(forStatement.update);
        this.body(forStatement.body);
        this.leave(forStatement);
    }
    functionDeclaration(functionDeclaration) {
        if (functionDeclaration.id) {
            this.scope.hoist(functionDeclaration.id);
        }
        this.deferVisit(() => {
            this.enter(functionDeclaration);
            this.visitAll(functionDeclaration.params);
            this.blockStatement(functionDeclaration.body, false);
            this.leave(functionDeclaration);
        });
    }
    ifStatement(ifStatement) {
        this.visit(ifStatement.test);
        this.visit(ifStatement.consequent);
        this.visitNullable(ifStatement.alternate);
    }
    labeledStatement(labeledStatement) {
        this.visit(labeledStatement.body);
    }
    returnStatement(returnStatement) {
        this.visitNullable(returnStatement.argument);
    }
    switchStatement(switchStatement) {
        this.visit(switchStatement.discriminant);
        this.enter(switchStatement);
        this.visitAll(switchStatement.cases);
        this.leave(switchStatement);
    }
    throwStatement(throwStatement) {
        this.visit(throwStatement.argument);
    }
    tryStatement(tryStatement) {
        this.visit(tryStatement.block);
        this.visitNullable(tryStatement.handler);
        this.visitNullable(tryStatement.finalizer);
    }
    variableDeclaration(variableDeclaration) {
        if (variableDeclaration.kind === "var") {
            this.reportError(error_1.Error.CannotDeclareWithVar);
        }
        variableDeclaration.declarations.forEach(this.variableDeclarator.bind(this));
    }
    whileStatement(whileStatement) {
        this.visit(whileStatement.test);
        this.visit(whileStatement.body);
    }
    withStatement(_withStatement) {
        this.reportError(error_1.Error.ForbiddenSyntax);
    }
    classDeclaration(classDeclaration) {
        if (classDeclaration.superClass) {
            this.visit(classDeclaration.superClass);
        }
        this.scope.hoist(classDeclaration.id);
        this.enter(classDeclaration);
        this.classBody(classDeclaration.body);
        this.leave(classDeclaration);
    }
    exportAllDeclaration(_exportAllDeclaration) {
        this.reportError(error_1.Error.ModuleDisallowed);
    }
    exportDefaultDeclaration(_exportDefaultDeclaration) {
        this.reportError(error_1.Error.ModuleDisallowed);
    }
    exportNamedDeclaration(_exportNamedDeclaration) {
        this.reportError(error_1.Error.ModuleDisallowed);
    }
    forOfStatement(forOfStatement) {
        this.enter(forOfStatement);
        this.visit(forOfStatement.left);
        this.visit(forOfStatement.right);
        this.body(forOfStatement.body);
        this.leave(forOfStatement);
    }
    importDeclaration(_importDeclaration) {
        this.reportError(error_1.Error.ModuleDisallowed);
    }
    // =============================================
    // Start: all possible sub-types of `Expression`
    // =============================================
    arrayExpression(arrayExpression) {
        this.visitAll(arrayExpression.elements.filter(helpers_1.removeNullable));
    }
    assignmentExpression(assignmentExpression) {
        this.visit(assignmentExpression.left);
        this.visit(assignmentExpression.right);
    }
    binaryExpression(binaryExpression) {
        this.visit(binaryExpression.left);
        this.visit(binaryExpression.right);
    }
    callExpression(callExpression) {
        this.visit(callExpression.callee);
        this.visitAll(callExpression.arguments);
    }
    conditionalExpression(conditionalExpression) {
        this.visit(conditionalExpression.test);
        this.visit(conditionalExpression.consequent);
        this.visit(conditionalExpression.alternate);
    }
    functionExpression(functionExpression) {
        this.deferVisit(() => {
            this.enter(functionExpression);
            this.visitAll(functionExpression.params);
            this.blockStatement(functionExpression.body, false);
            this.leave(functionExpression);
        });
    }
    identifier(identifier) {
        if (this.isDefineMode) {
            this.scope.define(identifier.name);
        }
        else {
            this.scope.resolve(identifier);
        }
    }
    stringLiteral(_stringLiteral) { }
    numericLiteral(_numericLiteral) { }
    nullLiteral(_nullLiteral) { }
    booleanLiteral(_booleanLiteral) { }
    regExpLiteral(_regExpLiteral) { }
    logicalExpression(logicalExpression) {
        this.visit(logicalExpression.left);
        this.visit(logicalExpression.right);
    }
    memberExpression(memberExpression) {
        this.visit(memberExpression.object);
        if (memberExpression.computed) {
            this.visit(memberExpression.property);
        }
    }
    newExpression(newExpression) {
        this.visit(newExpression.callee);
        this.visitAll(newExpression.arguments);
    }
    objectExpression(objectExpression) {
        this.visitAll(objectExpression.properties);
    }
    sequenceExpression(sequenceExpression) {
        this.visitAll(sequenceExpression.expressions);
    }
    parenthesizedExpression(parenthesizedExpression) {
        this.visit(parenthesizedExpression);
    }
    thisExpression(_thisExpression) { }
    unaryExpression(unaryExpression) {
        return this.visit(unaryExpression.argument);
    }
    updateExpression(updateExpression) {
        return this.visit(updateExpression.argument);
    }
    arrowFunctionExpression(arrowFunctionExpression) {
        this.deferVisit(() => {
            this.enter(arrowFunctionExpression);
            this.visitAll(arrowFunctionExpression.params);
            this.body(arrowFunctionExpression.body);
            this.leave(arrowFunctionExpression);
        });
    }
    classExpression(classExpression) {
        if (classExpression.id) {
            this.scope.hoist(classExpression.id);
        }
        this.visitNullable(classExpression.superClass);
        this.classBody(classExpression.body);
    }
    metaProperty(metaProperty) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    super(_super) { }
    taggedTemplateExpression(taggedTemplateExpression) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    templateLiteral(templateLiteral) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    templateElement(templateElement) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    yieldExpression(yieldExpression) {
        return this.visitNullable(yieldExpression.argument);
    }
    awaitExpression(awaitExpression) {
        return this.visit(awaitExpression.argument);
    }
    import(_import) {
        this.reportError(error_1.Error.ModuleDisallowed);
    }
    bigIntLiteral(_bigIntLiteral) { }
    optionalMemberExpression(optionalMemberExpression) {
        this.visit(optionalMemberExpression.object);
        if (optionalMemberExpression.computed) {
            this.visit(optionalMemberExpression.property);
        }
    }
    optionalCallExpression(optionalCallExpression) {
        this.visit(optionalCallExpression.callee);
        this.visitAll(optionalCallExpression.arguments);
    }
    // =============================
    // Start: JSX related node types
    // =============================
    jsxIdentifier(_jsxIdentifier) {
        // TODO: check when this is useful.
    }
    jsxMemberExpression(jsxMemberExpression) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    jsxElement(jsxElement) {
        this.jsxOpeningElement(jsxElement.openingElement);
        this.visitAll(jsxElement.children);
    }
    jsxSpreadChild(jsxSpreadChild) {
        return this.visit(jsxSpreadChild.expression);
    }
    jsxFragment(jsxFragment) {
        return this.visitAll(jsxFragment.children);
    }
    jsxOpeningElement(jsxOpeningElement) {
        return this.visitAll(jsxOpeningElement.attributes);
    }
    jsxOpeningFragment(_jsxOpeningFragment) { }
    jsxClosingElement(_jsxClosingElement) { }
    jsxClosingFragment(_jsxClosingFragment) { }
    jsxAttribute(jsxAttribute) {
        this.visitNullable(jsxAttribute.value);
    }
    jsxSpreadAttribute(jsxSpreadAttribute) {
        return this.visit(jsxSpreadAttribute.argument);
    }
    jsxExpressionContainer(jsxExpressionContainer) {
        return this.visit(jsxExpressionContainer.expression);
    }
    jsxText(_jsxText) { }
    jsxEmptyExpression(_jsxEmptyExpression) { }
    // ===========================
    // End: JSX related node types
    // ===========================
    bindExpression(_bindExpression) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    pipelinePrimaryTopicReference(_pipelinePrimaryTopicReference) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    doExpression(_doExpression) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    recordExpression(_recordExpression) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    tupleExpression(_tupleExpression) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    decimalLiteral(_decimalLiteral) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    moduleExpression(_moduleExpression) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    restElement(restElement) {
        return this.visit(restElement.argument);
    }
    assignmentPattern(assignmentPattern) {
        this.visit(assignmentPattern.left);
        this.setDefineMode(false);
        this.visit(assignmentPattern.right);
        this.unsetDefineMode();
    }
    arrayPattern(arrayPattern) {
        this.visitAll(arrayPattern.elements.filter(helpers_1.removeNullable));
    }
    objectPattern(objectPattern) {
        this.visitAll(objectPattern.properties);
    }
    objectMethod(objectMethod) {
        if (objectMethod.computed) {
            this.visit(objectMethod.key);
        }
        this.deferVisit(() => {
            this.enter(objectMethod);
            this.visitAll(objectMethod.params);
            this.blockStatement(objectMethod.body, false);
            this.leave(objectMethod);
        });
    }
    objectProperty(objectProperty) {
        if (objectProperty.computed) {
            this.setDefineMode(false);
            this.visit(objectProperty.key);
            this.unsetDefineMode();
        }
        this.visit(objectProperty.value);
    }
    // Misc
    catchClause(catchClause) {
        this.enter(catchClause);
        this.visitNullable(catchClause.param);
        this.blockStatement(catchClause.body);
        this.leave(catchClause);
    }
    /**
     * Let’s discuss the type of `declarator.id` is `LVal`.
     *
     * - Identifier: e.g. `y` in `let x = y`.
     * - MemberExpression: impossible.
     * - RestElement: impossible.
     * - AssignmentPattern: impossible.
     * - ArrayPattern: e.g. let [x, ..xs] = a;
     * - ObjectPattern: e.g. let { x, y = z, y: { t } } = a;
     * - TSParameterProperty: unsupported.
     *
     * @param declarator the `VariableDeclarator` to visit
     */
    variableDeclarator(declarator) {
        this.setDefineMode(true);
        this.visit(declarator.id);
        this.unsetDefineMode();
        this.visitNullable(declarator.init);
    }
    /**
     * E.g. `call(?, test)`.
     * @param _argumentPlaceholder the placeholder operator `?`
     */
    argumentPlaceholder(_argumentPlaceholder) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    jsxNamespacedName(_jsxNamespacedName) {
        this.reportError(error_1.Error.UnimplementedJSX);
    }
    spreadElement(spreadElement) {
        return this.visit(spreadElement.argument);
    }
    v8IntrinsicIdentifier(_v8IntrinsicIdentifier) {
        this.reportError(error_1.Error.ExpressionDisallowed);
    }
    privateName(_privateName) { }
    noop(_noop) { }
    classMethod(classMethod) {
        if (classMethod.computed) {
            this.visit(classMethod.key);
        }
        this.enter(classMethod);
        this.visitAll(classMethod.params);
        this.blockStatement(classMethod.body);
        this.leave(classMethod);
    }
    classBody(classBody) {
        classBody.body.forEach(this.visit.bind(this));
    }
    classPrivateMethod(_classPrivateMethod) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    /**
     * Stage 3: https://github.com/tc39/proposal-class-static-block
     * @param staticBlock the `StaticBlock` to visit.
     */
    staticBlock(staticBlock) {
        this.visitAll(staticBlock.body);
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    classPrivateProperty(_classPrivateProperty) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    classProperty(_classProperty) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    decorator(_decorator) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    directive(_directive) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    directiveLiteral(_directiveLiteral) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    exportSpecifier(_exportSpecifier) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    importDefaultSpecifier(_importDefaultSpecifier) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    importNamespaceSpecifier(_importNamespaceSpecifier) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    importSpecifier(_importSpecifier) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    exportNamespaceSpecifier(_exportNamespaceSpecifier) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    exportDefaultSpecifier(_exportDefaultSpecifier) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    importAttribute(_importAttribute) {
        this.reportError(error_1.Error.NotImplementedYet);
    }
    /**
     * See https://github.com/tc39/proposal-hashbang
     * @param _interpreterDirective the node
     */
    interpreterDirective(_interpreterDirective) { }
    pipelineBareFunction(pipelineBareFunction) {
        this.visit(pipelineBareFunction);
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    pipelineTopicExpression(pipelineTopicExpression) {
        this.visit(pipelineTopicExpression);
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    placeholder(_placeholder) {
        this.reportError(error_1.Error.ExperimentalSyntax);
    }
    switchCase(switchCase) {
        this.visitNullable(switchCase.test);
        this.visitAll(switchCase.consequent);
    }
    // ============================
    // Start: some deprecated nodes
    // ============================
    /** @deprecated */
    spreadProperty(_spreadProperty) {
        this.reportError(error_1.Error.Deprecated);
    }
    /** @deprecated */
    restProperty(_restProperty) {
        this.reportError(error_1.Error.Deprecated);
    }
    /** @deprecated */
    numberLiteral(_numberLiteral) {
        this.reportError(error_1.Error.Deprecated);
    }
    /** @deprecated */
    regexLiteral(_regexLiteral) {
        this.reportError(error_1.Error.Deprecated);
    }
}
exports.FreeVariables = FreeVariables;
